"""Cynthia language parser definition."""

from typing import Any, List

from cynthia.tools.lexer import Lexer, Token, TokenType
from cynthia.tools.nodes import (
    AssignmentNode, BinaryNode, FunctionNode, Node, NumberNode, ProgramNode, UnaryNode, VariableNode,
)


class Parser(object):
    """The parser creates the AST for a given token sequence."""

    def __init__(self, lexer: Lexer = None):
        self.lexer = lexer if lexer else Lexer()

    @property
    def current_token(self) -> Token:
        return self._get_token(offset=0)

    @property
    def next_token(self) -> Token:
        return self._get_token(offset=1)

    def parse(self, to_parse: str | List[Token]) -> ProgramNode:
        self._set_tokens(to_parse)
        try:
            return self.program()
        except Exception:
            raise SyntaxError('Syntax error.')

    def program(self) -> ProgramNode:
        """Get the root node of the program.

        program: statement_list EOF

        Returns:
            ProgramNode: root of the program
        """
        statements = self.statement_list()
        self.eat(TokenType.EOF)
        return ProgramNode(statements=statements)

    def statement_list(self) -> List[Node]:
        """Get each statement of the program.

        statement_list: statement | statement EOL statement_list

        Returns:
            List[StatementNode]: List of statements.
        """
        first_statement = self.statement()
        statements = [first_statement] if first_statement else []

        if self.current_token.type == TokenType.EOL:
            self.eat_current()
            statements.extend(self.statement_list())

        return statements

    def statement(self) -> Node:
        """Get a single statement.

        statement: assignment_statement | expr | EMPTY

        Returns:
            Node: Statement node.
        """
        if self._is_assignment():
            return self.assignment_statement()
        return self.expr()

    def assignment_statement(self) -> AssignmentNode:
        """Get assignment statement.

        assignment_statement: VARIABLE ASSIGN expr

        Returns:
            AssignmentNode: Assignment.
        """
        variable_node = VariableNode(token=self.eat(TokenType.ID))
        self.eat(TokenType.ASSIGN)
        return AssignmentNode(
            variable=variable_node,
            expression=self.expr(),
        )

    def expr(self) -> Node:
        """Get an expression.

        expr: term ((PLUS | MINUS) term)*

        Returns:
            Node: expression
        """
        node = self.term()
        while self.current_token.type in {TokenType.PLUS, TokenType.MINUS}:
            token = self.eat_current()
            if right := self.term():
                node = BinaryNode(left=node, token=token, right=right)
            else:
                self.error()
        return node

    def term(self) -> Node:
        """Get a term.

        term: factor ((MUL | DIV) factor)*

        Returns:
            Node: expression
        """
        node = self.factor()
        while self.current_token.type in {TokenType.ASTERISK, TokenType.FORWARD_SLASH}:
            token = self.eat_current()
            if right := self.factor():
                node = BinaryNode(left=node, token=token, right=right)
            else:
                self.error()
        return node

    def factor(self) -> Node:
        """Get a factor.

        factor
            : NUMBER
            | VARIABLE
            | L_PAREN expr R_PAREN
            | (PLUS|MINUS) expr
            | function_call
            ;

        Returns:
            Node: factor
        """
        token = self.current_token
        node = None

        match(token.type):
            case TokenType.INTEGER:
                node = NumberNode(token=self.eat_current())
            case TokenType.ID:
                if self.next_token.type == TokenType.L_PAREN:
                    node = self.function_call()
                else:
                    node = VariableNode(token=self.eat_current())
            case TokenType.L_PAREN:
                self.eat_current()
                node = self.expr()
                self.eat(TokenType.R_PAREN)
            case TokenType.MINUS | TokenType.PLUS:
                self.eat_current()
                node = UnaryNode(child=self.factor(), token=token)

        return node

    def function_call(self) -> FunctionNode:
        """Get a function call.

        function_call: ID L_PAREN args R_PAREN

        Returns:
            FunctionNode: function call.
        """
        id_token = self.eat_current()
        self.eat(TokenType.L_PAREN)
        args = self.args()
        self.eat(TokenType.R_PAREN)
        return FunctionNode(token=id_token, args=args)

    def args(self) -> List[Node]:
        """Get arguments.

        args: expr (COMMA expr)*

        Returns:
            List[Node]: list of arguments.
        """
        args = [self.expr()]
        while self.current_token.type == TokenType.COMMA:
            self.eat_current()
            if extra_arg := self.expr():
                args.append(extra_arg)
            else:
                self.error()
        return args

    def eat_current(self):
        if self.current_token:
            return self.eat(self.current_token.type)
        return None

    def eat(self, token_type: TokenType):
        if self.current_token.type == token_type:
            eaten_token = self.current_token
            self._pos += 1
            return eaten_token
        raise ValueError(
            'Tried to eat {expected_type} however next token was a {actual_type}'.format(
                expected_type=token_type,
                actual_type=self.current_token.type,
            ),
        )

    def _set_tokens(self, to_parse: str | List[Token]):
        self._pos = 0
        if isinstance(to_parse, str):
            self._tokens = self.lexer.get_tokens(to_parse)
        elif self._is_token_list(to_parse):
            self._tokens = to_parse
        else:
            raise ValueError('Only strings or list of tokens are supported.')

    def _is_assignment(self) -> bool:
        return (
            self.current_token.type == TokenType.ID and
            self.next_token.type == TokenType.ASSIGN
        )

    def _is_token_list(self, value_to_check: Any) -> bool:
        return isinstance(value_to_check, list) and self._each_element_is_token(value_to_check)

    def _each_element_is_token(self, list_to_check) -> bool:
        for token in list_to_check:
            if not isinstance(token, Token):
                return False
        return True

    def _get_token(self, offset: int) -> Token:
        try:
            return self._tokens[self._pos + offset]
        except (AttributeError, IndexError):
            return None



"""General token definition."""

from enum import Enum
from typing import Any


class Token(object):
    """Class for representing any input as token."""

    def __init__(self, type: Enum, value: Any):
        """Initialize token."""
        self.type = type
        self.value = value

    def __str__(self):
        """String representation of the class instance."""
        return 'Token({type}, {value})'.format(
            type=self.type,
            value=repr(self.value)
        )

    def __repr__(self):
        """Show string representation."""
        return self.__str__()



"""Utility functions for the lexer module."""

from cynthia.tools.lexer.supported_tokens import TokenType, white_space_tokens


def is_white_space(token_type: TokenType) -> bool:
    return token_type in white_space_tokens



"""AST node definitions for the Cynthia language."""

from abc import ABC
from dataclasses import dataclass
from decimal import Decimal
from typing import List

from cynthia.tools.lexer import Token, TokenType


@dataclass(kw_only=True)
class Node(ABC):
    """Class representing an AST node."""

    token: Token = None


@dataclass(kw_only=True)
class VariableNode(Node):
    """Variable node representation."""

    name: str | None = None

    def __post_init__(self):
        if self.name is None:
            self.name = self.token.value  # noqa: WPS601


@dataclass
class ProgramNode(Node):
    """ProgramNode representation.

    This node will be the root of the AST.
    """

    statements: List[Node]


@dataclass(kw_only=True)
class BinaryNode(Node):
    """BinaryNode representation."""

    left: Node
    right: Node


@dataclass(kw_only=True)
class AssignmentNode(Node):
    """Separate node for assignments."""

    token: Token = Token(TokenType.ASSIGN, '=')
    variable: VariableNode  # noqa: WPS110
    expression: Node


@dataclass(kw_only=True)
class UnaryNode(Node):
    """UnaryNode representation."""

    child: Node


@dataclass(kw_only=True)
class FunctionNode(VariableNode):
    """Function node representation."""

    args: list[Node]


@dataclass(kw_only=True)
class NumberNode(Node):
    """Number node representation."""

    amount: Decimal | None = None

    def __post_init__(self):
        """Set value."""
        if self.amount is None:
            self.amount = Decimal(self.token.value)  # noqa: WPS601



"""Cynthia language interpreter definition."""

import re
from io import TextIOBase

from cynthia.tools.interpreter.exceptions import (
    FunctionDoesNotExistError, VariableDoesNotExitError, VisitorDoesNotExistError,
)
from cynthia.tools.lexer.supported_tokens import TokenType
from cynthia.tools.nodes import (
    AssignmentNode, BinaryNode, FunctionNode, Node, NumberNode, ProgramNode, UnaryNode, VariableNode,
)
from cynthia.tools.parser import Parser


class NodeVisitor(object):
    """Visitor pattern for nodes."""

    def visit(self, node: Node):
        method_name = 'visit_{node_name}'.format(
            node_name=self._pascal_to_snake(type(node).__name__),
        )
        visitor = getattr(self, method_name, self.generic_visit)
        return visitor(node)

    def generic_visit(self, node):
        raise VisitorDoesNotExistError(node=node)

    def _pascal_to_snake(self, pascal_name: str) -> str:
        return re.sub('(?<!^)(?=[A-Z])', '_', pascal_name).lower()


class Interpreter(NodeVisitor):
    """Interpreter will evaluate a given AST."""

    supported_file_extensions = ['cynthia']

    binary_logics = {
        TokenType.PLUS: lambda left, right: left + right,
        TokenType.MINUS: lambda left, right: left - right,
        TokenType.ASTERISK: lambda left, right: left * right,
        TokenType.FORWARD_SLASH: lambda left, right: left / right,
    }

    unary_logics = {
        TokenType.PLUS: lambda child: child,
        TokenType.MINUS: lambda child: -child,
    }

    core_functions = {
        'max': lambda *args: max(*args),
        'min': lambda *args: min(*args),
        'print': lambda *args: print(*args),
    }

    def __init__(self, parser: Parser = None, global_variables: dict = None):
        self.parser = parser if parser else Parser()
        self.global_variables = global_variables if global_variables else {}

    def visit_program_node(self, node: ProgramNode):
        for statement in node.statements:
            self.visit(statement)

    def visit_binary_node(self, node: BinaryNode):
        if binary_logic := self.binary_logics.get(node.token.type):
            return binary_logic(self.visit(node.left), self.visit(node.right))
        return None

    def visit_assignment_node(self, node: AssignmentNode):
        self.global_variables[node.variable.name] = self.visit(node.expression)

    def visit_unary_node(self, node: UnaryNode):
        if unary_logic := self.unary_logics.get(node.token.type):
            return unary_logic(node.child)
        return None

    def visit_number_node(self, number: NumberNode):
        return number.amount

    def visit_variable_node(self, variable_node: VariableNode):
        try:
            return self.global_variables[variable_node.name]
        except KeyError:
            raise VariableDoesNotExitError(f' Variable {variable_node.name} does not exist.')

    def visit_function_node(self, function_node: FunctionNode):
        args = [self.visit(arg) for arg in function_node.args]
        if core_function := self.core_functions.get(function_node.token.value):
            return core_function(*args)
        raise FunctionDoesNotExistError(f' Function {function_node.name} does not exist.')

    def interpret(self, to_interpret: str | Node | TextIOBase):
        if isinstance(to_interpret, TextIOBase):
            self._check_file_extension(to_interpret)
            to_interpret = to_interpret.read()
        if isinstance(to_interpret, str):
            to_interpret = self.parser.parse(to_interpret)
        return self.visit(to_interpret)

    def shell_interpret(self, to_interpret: str):
        print_interpret = f'print({to_interpret})\n'
        try:
            to_interpret = self.parser.parse(print_interpret)
        except Exception:
            to_interpret = self.parser.parse(to_interpret)
        self.interpret(to_interpret)

    def __getitem__(self, variable_name):
        return self.global_variables.get(variable_name)

    def _check_file_extension(self, input_file: TextIOBase):
        if input_file.name.split('.')[-1] not in self.supported_file_extensions:
            raise TypeError(
                'Allowed file extensions are: {extensions}'.format(
                    extensions=self.supported_file_extensions,
                ),
            )



"""Supported tokens for the Cynthia language."""

from enum import Enum


class TokenType(Enum):
    """Supported token types."""

    INTEGER = 'INTEGER'
    PLUS = 'PLUS'
    MINUS = 'MINUS'
    ASTERISK = 'ASTERISK'
    FORWARD_SLASH = 'FORWARD_SLASH'
    L_PAREN = 'L_PAREN'
    R_PAREN = 'R_PAREN'
    ASSIGN = 'ASSIGN'
    ID = 'ID'
    EOF = 'EOF'
    EOL = 'EOL'
    SPACE = 'SPACE'
    COMMA = 'COMMA'
    COMMENT = 'COMMENT'


white_space_tokens = [
    TokenType.SPACE,
]



"""Logic definitions for token creation."""

import logging
from abc import ABC, abstractmethod
from enum import Enum
from functools import wraps
from string import ascii_letters, digits, printable

from cynthia.tools.lexer.tokens import Token


def set_text_and_pos(handle_input):
    @wraps(handle_input)
    def wrapper(self, input_txt: str, pos: int):
        self._text = input_txt
        self._pos = pos
        return handle_input(self, input_txt, pos)
    return wrapper


class TokenLogic(ABC):
    """Class representing token creating logic.

    TokenLogics will handle token creation for given text with a given
    position.

    The `handle_input` method should return the created Token and the number
    of consumed characters from the text.

    The TokenLogic can not create a Token it should return None.
    """

    def __init__(self, token_type: Enum):
        self.token_type = token_type

    @abstractmethod
    def handle_input(self, input_txt: str, pos: int) -> Token | None:
        """Handle input for with given position.

        Args:
            input_txt (str): Complete input text.
            pos (int): Current position.

        Returns:
            Token | None: Token if created else None.
        """

    @property
    def current_char(self) -> str:
        try:
            return self._text[self._pos]
        except (IndexError):
            return None
        except AttributeError:
            logging.warning('_text was not set! Use the `set_text_and_pos` decorator.')

    def advance(self):
        try:
            self._pos += 1
        except AttributeError:
            logging.warning('_pos was not set! Use the `set_text_and_pos` decorator.')


class TextMatcherTokenLogic(TokenLogic):
    """Token logic for matching entire texts."""

    def __init__(self, *args, text_to_match: str, **kwargs):
        super().__init__(*args, **kwargs)
        self._text_to_match = text_to_match

    @set_text_and_pos
    def handle_input(self, input_txt: str, pos: int) -> Token:
        received_text = self._get_sub_text(input_txt, pos)
        if received_text and received_text == self._text_to_match:
            return Token(self.token_type, received_text)
        return None

    def _get_sub_text(self, text: str, pos: int) -> str:
        slice_until = pos + len(self._text_to_match)
        try:
            return text[pos:slice_until]
        except IndexError:
            return None


class PatternMatcherTokenLogic(TokenLogic):
    """Token logic for matching patterns."""

    supported_first_chars: str
    supported_chars: str

    @set_text_and_pos
    def handle_input(self, input_txt: str, pos: int) -> Token:
        text_matched = ''

        if self.current_char is None or self.current_char not in self.supported_first_chars:
            return None

        while self.current_char and self.current_char in self.supported_chars:
            text_matched += self.current_char
            self.advance()

        return Token(self.token_type, text_matched) if text_matched else None


class IDNameTokenLogic(PatternMatcherTokenLogic):
    """Token logic for ID names.

    ID names will be used for variable names and function names.

    ID names can contain lowercase and uppercase letters digits and the
    underscore character. ID names first letter can not be a digit.
    """

    supported_first_chars = ascii_letters
    supported_chars = f'{ascii_letters}{digits}_'


class NumberTokenLogic(PatternMatcherTokenLogic):
    """Token logic for numbers."""

    supported_first_chars = digits
    supported_chars = digits

    def handle_input(self, input_txt: str, pos: int) -> Token:
        # get leading part of the number
        if token := super().handle_input(input_txt, pos):
            if self.current_char == '.':
                if trailing_token := super().handle_input(input_txt, pos + 1):
                    token.value += f'.{trailing_token.value}'
                else:
                    token = None
        return token


class CommentTokenLogic(PatternMatcherTokenLogic):
    """Token logic for comments."""

    supported_first_chars = '#'
    supported_chars = printable.replace('\n', '')



"""Cynthia language lexer definition."""

from typing import List

from cynthia.tools.lexer.lexer_utils import is_white_space
from cynthia.tools.lexer.logics import (
    CommentTokenLogic, IDNameTokenLogic, NumberTokenLogic, TextMatcherTokenLogic, TokenLogic,
)
from cynthia.tools.lexer.supported_tokens import TokenType
from cynthia.tools.lexer.tokens import Token


class UnsupportedTokenError(Exception):
    """Raised when the input wasn't fully processed."""


class Lexer(object):
    """The lexer's job is to make a token sequence for a given text input.

    Tokens will be created via the `token_create_logics`.

    The lexer will iterate through the list and the first logic that does not
    return None will be used.
    """

    token_create_logics: List[TokenLogic] = [
        TextMatcherTokenLogic(TokenType.PLUS, text_to_match='+'),
        TextMatcherTokenLogic(TokenType.MINUS, text_to_match='-'),
        TextMatcherTokenLogic(TokenType.ASTERISK, text_to_match='*'),
        TextMatcherTokenLogic(TokenType.FORWARD_SLASH, text_to_match='/'),
        TextMatcherTokenLogic(TokenType.ASSIGN, text_to_match='='),
        TextMatcherTokenLogic(TokenType.L_PAREN, text_to_match='('),
        TextMatcherTokenLogic(TokenType.R_PAREN, text_to_match=')'),
        TextMatcherTokenLogic(TokenType.COMMA, text_to_match=','),
        TextMatcherTokenLogic(TokenType.SPACE, text_to_match=' '),
        TextMatcherTokenLogic(TokenType.EOL, text_to_match='\n'),
        NumberTokenLogic(TokenType.INTEGER),
        IDNameTokenLogic(TokenType.ID),
        CommentTokenLogic(TokenType.COMMENT),
    ]

    def __init__(self, include_ws: bool = False):
        self.include_ws = include_ws

    def get_tokens(self, text: str) -> List[Token]:
        position = 0
        tokens = []

        token = self.get_next_token(text, position)
        while token:
            if self._check_white_space(token) and token.type != TokenType.COMMENT:
                tokens.append(token)
            position += len(token.value)
            token = self.get_next_token(text, position)

        if position + 1 < len(text):
            raise UnsupportedTokenError(f'Unsupported input starting with {text[position]}.')

        tokens.append(Token(TokenType.EOF, None))
        return tokens

    def get_next_token(self, text: str, position: int):
        for logic in self.token_create_logics:
            if token := logic.handle_input(text, position):
                return token
        return None

    def _check_white_space(self, token):
        return self.include_ws or not is_white_space(token.type)



"""Shell application for the Cynthia language."""

from cynthia.tools.interpreter.interpreter import Interpreter


def run_shell():
    interpreter = Interpreter()
    while True:
        text = input('cynthia> ')

        if text == 'exit()':
            break

        try:
            interpreter.shell_interpret(text)
        except Exception as exc:
            print(exc)



"""Interpreter related exceptions."""


class InterpreterError(Exception):
    """Base class for interpreter related exceptions."""


class VisitorDoesNotExistError(InterpreterError):
    """Exception when no visitor does not exist."""


class VariableDoesNotExitError(InterpreterError):
    """Exception when variable does not exist."""


class FunctionDoesNotExistError(InterpreterError):
    """Exception when function does not exist."""



"""Main module definition."""

import argparse

from cynthia.tools.interpreter.interpreter import Interpreter
from cynthia.tools.shell import run_shell


def main():
    """Run shell or interpret input file."""
    parser = argparse.ArgumentParser()
    parser.add_argument('file', nargs='?', type=argparse.FileType('r'), default=None)
    args = parser.parse_args()

    if args.file is None:
        run_shell()
    else:
        interpreter = Interpreter()
        try:
            interpreter.interpret(args.file)
        except Exception as exc:
            print(exc)


if __name__ == '__main__':
    main()
